# -*- coding: utf-8 -*-
# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

import os
from copy import deepcopy
from time import time
from typing import Dict, List, NoReturn, Tuple, Union

import numpy as np
import pandas as pd
from jurity.recommenders import CombinedMetrics, BinaryRecoMetrics, RankingRecoMetrics
from sklearn.model_selection import GroupKFold
from mabwiser.utils import check_true, Arm

from mab2rec.rec import BanditRecommender
from mab2rec.utils import Constants
from mab2rec.utils import explode_recommendations, load_data, load_response_data, merge_user_features
from mab2rec.utils import load_pickle, save_pickle


def train(recommender: BanditRecommender,
          data: Union[str, pd.DataFrame],
          user_features: Union[str, pd.DataFrame] = None,
          user_features_list: Union[str, List[str]] = None,
          user_features_dtypes: Union[str, Dict] = None,
          item_features: Union[str, pd.DataFrame] = None,
          item_list: Union[str, List[Arm]] = None,
          item_eligibility: Union[str, pd.DataFrame] = None,
          warm_start: bool = False,
          warm_start_distance: float = None,
          user_id_col: str = Constants.user_id,
          item_id_col: str = Constants.item_id,
          response_col: str = Constants.response,
          batch_size: int = 100000,
          save_file: Union[str, bool] = None) -> NoReturn:
    """
    Trains Recommender.

    Parameters
    ----------
    recommender : BanditRecommender
        The recommender algorithm to be trained.
        The recommender object is updated in-place.
    data : Union[str, pd.DataFrame]
        Training data.
        Data should have a row for each training sample (user_id, item_id, response).
        Column names should be consistent with user_id_col, item_id_col and response_col arguments.
        CSV format with file header or Data Frame.
    user_features : Union[str, pd.DataFrame], default=None
        User features containing features for each user_id.
        Each row should include user_id and list of features (user_id, u_1, u_2, ..., u_p).
        CSV format with file header or Data Frame.
    user_features_list : Union[str, List[str]], default=None
        List of user features to use.
        Must be a subset of features in (u_1, u_2, ... u_p).
        If None, all the features in user_features are used.
        CSV format with file header or List.
    user_features_dtypes: Union[str, Dict], default=None
        Data type for each user feature.
        Maps each user feature name to valid data type.
        If none, no data type casting is done upon load and data types or inferred by Pandas library.
        JSON format or Dictionary.
    item_features : Union[str, pd.DataFrame], default=None
        Item features file containing features for each item_id.
        Each row should include item_id and list of features (item_id, i_1, i_2, .... i_q).
        CSV format with file header or Data Frame.
    item_list : Union[str, List[Arm]], default=None
        List of items to train.
        If None, all the items in data are used.
        CSV format with file header or List.
    item_eligibility: Union[str, pd.DataFrame], default=None
        Items each user is eligible for.
        Not used during training.
        CSV format with file header or Data Frame.
    warm_start : bool, default=False
        Whether to warm start untrained (cold) arms after training or not.
    warm_start_distance : float, default=None
        Warm start distance quantile.
        Value between 0 and 1 used to determine if an item can be warm started or not using closest item.
        All cold items will be warm started if 1 and none will be warm started if 0.
        Must be specified if warm_start=True.
    user_id_col : str, default=Constants.user_id
        User id column name.
    item_id_col : str, default=Constants.item_id
        Item id column name.
    response_col : str, default=Constants.response
        Response column name.
    batch_size : str, default=100000
        Batch size used for chunking data.
    save_file : Union[str, bool], default=None
        File name to save recommender pickle.
        If None, recommender is not saved to file.

    Returns
    -------
    No Return.
    """
    _validate_recommender(recommender)
    _validate_common_args(data, user_features, user_features_list, user_features_dtypes, item_features, item_list,
                          item_eligibility, warm_start, warm_start_distance, user_id_col, item_id_col,
                          response_col, batch_size)
    _validate_save(save_file)

    # Import data
    train_data_df, item_list, user_features_df, \
        item_to_features, _ = load_data(data=data,
                                        user_features=user_features,
                                        user_features_list=user_features_list,
                                        user_features_dtypes=user_features_dtypes,
                                        item_features=item_features,
                                        item_list=item_list,
                                        item_eligibility=item_eligibility,
                                        user_id_col=user_id_col,
                                        item_id_col=item_id_col,
                                        response_col=response_col)

    # Initialize and set arms of recommender
    recommender.set_arms(item_list)

    # Loop through the data in batches and fit recommender
    num_batches = max(1, len(train_data_df) // batch_size)
    for df in np.array_split(train_data_df, num_batches):
        if recommender.mab.is_contextual:
            check_true(user_features_df is not None, ValueError("User features are required for contextual bandits."))
            feature_cols = [c for c in user_features_df.columns if c != user_id_col]
            df = merge_user_features(pd.DataFrame(df), user_features_df, user_id_col)
            if recommender.mab._is_initial_fit:
                recommender.partial_fit(df[item_id_col], df[response_col], df[feature_cols])
            else:
                recommender.fit(df[item_id_col], df[response_col], df[feature_cols])
        else:
            if recommender.mab._is_initial_fit:
                recommender.partial_fit(df[item_id_col], df[response_col])
            else:
                recommender.fit(df[item_id_col], df[response_col])

    # Warm start
    if warm_start:
        recommender.warm_start(item_to_features, warm_start_distance)

    # Save file
    if save_file is not None:
        if isinstance(save_file, str):
            os.makedirs(os.path.dirname(save_file), exist_ok=True)
            save_pickle(recommender, save_file)
        elif save_file:
            save_pickle(recommender, "recommender.pkl")


def score(recommender: Union[str, BanditRecommender],
          data: Union[str, pd.DataFrame],
          user_features: Union[str, pd.DataFrame] = None,
          user_features_list: Union[str, List[str]] = None,
          user_features_dtypes: Union[str, Dict] = None,
          item_features: Union[str, pd.DataFrame] = None,
          item_list: Union[str, List[Arm]] = None,
          item_eligibility: Union[str, pd.DataFrame] = None,
          warm_start: bool = False,
          warm_start_distance: float = None,
          user_id_col: str = Constants.user_id,
          item_id_col: str = Constants.item_id,
          response_col: str = Constants.response,
          batch_size: int = 100000,
          save_file: Union[str, bool] = None) -> pd.DataFrame:
    """
    Score Recommender.

    Generates top-k recommendations for users in given data.

    Parameters
    ----------
    recommender : Union[str, BanditRecommender]
        The recommender algorithm to be scored.
        Could be an instantiated BanditRecommender or file path of serialized recommender in pickle file.
    data : Union[str, pd.DataFrame]
        Training data.
        Data should have a row for each training sample (user_id, item_id, response).
        Column names should be consistent with user_id_col, item_id_col and response_col arguments.
        CSV format with file header or Data Frame.
    user_features : Union[str, pd.DataFrame], default=None
        User features containing features for each user_id.
        Each row should include user_id and list of features (user_id, u_1, u_2, ..., u_p).
        CSV format with file header or Data Frame.
    user_features_list : Union[str, List[str]], default=None
        List of user features to use.
        Must be a subset of features in (u_1, u_2, ... u_p).
        If None, all the features in user_features are used.
        CSV format with file header or List.
    user_features_dtypes: Union[str, Dict], default=None
        Data type for each user feature.
        Maps each user feature name to valid data type.
        If none, no data type casting is done upon load and data types or inferred by Pandas library.
        JSON format or Dictionary.
    item_features : Union[str, pd.DataFrame], default=None
        Item features file containing features for each item_id.
        Each row should include item_id and list of features (item_id, i_1, i_2, .... i_q).
        CSV format with file header or Data Frame.
    item_list : Union[str, List[Arm]], default=None
        List of items to train.
        If None, all the items in data are used.
        CSV format with file header or List.
    item_eligibility: Union[str, pd.DataFrame], default=None
        Items each user is eligible for.
        Used to generate excluded_arms lists.
        If None, all the items can be evaluated for recommendation for each user.
        CSV format with file header or Data Frame.
    warm_start : bool, default=False
        Whether to warm start untrained (cold) arms after training or not.
    warm_start_distance : float, default=None
        Warm start distance quantile.
        Value between 0 and 1 used to determine if an item can be warm started or not using closest item.
        All cold items will be warm started if 1 and none will be warm started if 0.
        Must be specified if warm_start=True.
    user_id_col : str, default=Constants.user_id
        User id column name.
    item_id_col : str, default=Constants.item_id
        Item id column name.
    response_col : str, default=Constants.response
        Response column name.
    batch_size : str, default=100000
        Batch size used for chunking data.
    save_file : str, default=None
        File name to save recommender pickle.
        If None, recommender is not saved to file.

    Returns
    -------
    Scored recommendations.
    """
    # Load recommender
    if isinstance(recommender, str):
        recommender = load_pickle(recommender)
    _validate_recommender(recommender, is_fit=True)
    _validate_common_args(data, user_features, user_features_list, user_features_dtypes, item_features,
                          item_list, item_eligibility, warm_start, warm_start_distance, user_id_col, item_id_col,
                          response_col, batch_size)
    _validate_save(save_file)

    # Import data
    test_data_df, item_list_out, user_features_df, \
        item_to_features, excluded_df = load_data(data=data,
                                                  user_features=user_features,
                                                  user_features_list=user_features_list,
                                                  user_features_dtypes=user_features_dtypes,
                                                  item_features=item_features,
                                                  item_list=item_list,
                                                  item_eligibility=item_eligibility,
                                                  user_id_col=user_id_col,
                                                  item_id_col=item_id_col,
                                                  response_col=response_col)
    # Set arms to recommender
    if item_list is not None:
        recommender.set_arms(item_list_out)

    # Warm start
    if warm_start:
        recommender.warm_start(item_to_features, warm_start_distance)

    # Loop through users in batches and get recommendations
    users = test_data_df[user_id_col].unique().tolist()
    recommendations = []
    scores = []
    num_batches = max(1, len(users) // batch_size)
    for users_of_batch in np.array_split(users, num_batches):

        # Data frame of users to score
        df = pd.DataFrame({user_id_col: users_of_batch})

        # Merge user features and get feature column names
        if user_features_df is not None:
            df = merge_user_features(df, user_features_df, user_id_col)
            feature_cols = [c for c in user_features_df.columns if c != user_id_col]
            contexts = df[feature_cols].fillna(0)
        else:
            contexts = [[]] * len(df)

        # Merge excluded item list
        if item_eligibility is not None:
            df = df.merge(excluded_df, how='left', on=user_id_col)
            excluded_arms_batch = df[item_id_col].tolist()
        else:
            excluded_arms_batch = None

        # Get recommendations
        recs_of_batch, scores_of_batch = recommender.recommend(contexts, excluded_arms_batch, return_scores=True)
        recommendations += recs_of_batch
        scores += scores_of_batch

    # Convert recommendations to data frame
    df = pd.DataFrame(users, columns=[user_id_col])
    df[item_id_col] = recommendations
    df[Constants.score] = scores
    df = explode_recommendations(df, user_id_col, [item_id_col, Constants.score])

    # Save to csv
    if save_file is not None:
        if isinstance(save_file, str):
            df.to_csv(save_file, index=False)
        elif save_file:
            df.to_csv("results.csv", index=False)

    return df


def benchmark(recommenders: Dict[str, BanditRecommender],
              metrics: List[Union[BinaryRecoMetrics, RankingRecoMetrics]],
              train_data: Union[str, pd.DataFrame],
              test_data: Union[str, pd.DataFrame] = None,
              cv: int = None,
              user_features: Union[str, pd.DataFrame] = None,
              user_features_list: Union[str, List[str]] = None,
              user_features_dtypes: Union[str, Dict] = None,
              item_features: Union[str, pd.DataFrame] = None,
              item_list: Union[str, List[Arm]] = None,
              item_eligibility: Union[str, pd.DataFrame] = None,
              warm_start: bool = False,
              warm_start_distance: float = None,
              user_id_col: str = Constants.user_id,
              item_id_col: str = Constants.item_id,
              response_col: str = Constants.response,
              batch_size: int = 100000,
              verbose: bool = False) -> Union[Tuple[Dict[str, pd.DataFrame], Dict[str, Dict[str, float]]],
                                              Tuple[List[Dict[str, pd.DataFrame]],
                                                    List[Dict[str, Dict[str, float]]]]]:

    """
    Benchmark Recommenders.

    Benchmark a given set of recommender algorithms by training, scoring and evaluating each algorithm
    If using cross-validation (cv) it benchmarks the algorithms on cv-many folds from the train data,
    otherwise it trains on the train data and evaluates on the test data.

    Parameters
    ----------
    recommenders : Dict[str, BanditRecommender]
        The recommender algorithms to be benchmarked.
        Dictionary with names (key) and recommender algorithms (value).
    metrics : List[Union[BinaryRecoMetrics, RankingRecoMetrics]]
        List of metrics used to evaluate recommendations.
    train_data : Union[str, pd.DataFrame]
        Training data used to train recommenders.
        Data should have a row for each training sample (user_id, item_id, response).
        Column names should be consistent with user_id_col, item_id_col and response_col arguments.
        CSV format with file header or Data Frame.
    test_data : Union[str, pd.DataFrame]
        Test data used to generate recommendations.
        Data should have a row for each training sample (user_id, item_id, response).
        Column names should be consistent with user_id_col, item_id_col and response_col arguments.
        CSV format with file header or Data Frame.
    cv : int, default=None
        Number of folds in the train data to use for cross-fold validation.
        A grouped K-fold iterator is used to ensure that the same user is not contained in different folds.
        Test data must be None when using cv.
    user_features : Union[str, pd.DataFrame], default=None
        User features containing features for each user_id.
        Each row should include user_id and list of features (user_id, u_1, u_2, ..., u_p).
        CSV format with file header or Data Frame.
    user_features_list : Union[str, List[str]], default=None
        List of user features to use.
        Must be a subset of features in (u_1, u_2, ... u_p).
        If None, all the features in user_features are used.
        CSV format with file header or List.
    user_features_dtypes: Union[str, Dict], default=None
        Data type for each user feature.
        Maps each user feature name to valid data type.
        If none, no data type casting is done upon load and data types or inferred by Pandas library.
        JSON format or Dictionary.
    item_features : Union[str, pd.DataFrame], default=None
        Item features file containing features for each item_id.
        Each row should include item_id and list of features (item_id, i_1, i_2, .... i_q).
        CSV format with file header or Data Frame.
    item_list : Union[str, List[Arm]], default=None
        List of items to train.
        If None, all the items in data are used.
        CSV format with file header or List.
    item_eligibility: Union[str, pd.DataFrame], default=None
        Items each user is eligible for.
        Used to generate excluded_arms lists.
        If None, all the items can be evaluated for recommendation for each user.
        CSV format with file header or Data Frame.
    warm_start : bool, default=False
        Whether to warm start untrained (cold) arms after training or not.
    warm_start_distance : float, default=None
        Warm start distance quantile.
        Value between 0 and 1 used to determine if an item can be warm started or not using closest item.
        All cold items will be warm started if 1 and none will be warm started if 0.
        Must be specified if warm_start=True.
    user_id_col : str, default=Constants.user_id
        User id column name.
    item_id_col : str, default=Constants.item_id
        Item id column name.
    response_col : str, default=Constants.response
        Response column name.
    batch_size : str, default=100000
        Batch size used for chunking data.
    verbose : bool, default=False
        Whether to print progress status or not.

    Returns
    -------
    Tuple with recommendations and evaluation metrics for each algorithm.
    The tuple values are lists of dictionaries if cross-validation is used, representing the results on each fold,
    and individual dictionaries otherwise.
    """
    _validate_recommender(recommenders)
    _validate_common_args(train_data, user_features, user_features_list, user_features_dtypes, item_features,
                          item_list, item_eligibility, warm_start, warm_start_distance, user_id_col,
                          item_id_col, response_col, batch_size)
    _validate_bench(recommenders, metrics, train_data, test_data, cv)

    # Convert input arguments to dictionary
    args = locals()
    args.pop('cv')

    if cv is None:
        return _bench(**args)
    else:

        # Read data
        if isinstance(train_data, str):
            df = pd.read_csv(train_data)
        else:
            df = pd.DataFrame(train_data)

        # Initialize lists to store recommendation results and metrics for each fold
        recommendations_list = []
        metrics_list = []

        # Split data into cv folds and run benchmark
        group_kfold = GroupKFold(n_splits=cv)
        i = 1
        for train_index, test_index in group_kfold.split(df, groups=df[user_id_col]):

            if verbose:
                print(f'CV Fold = {i} \n')

            # Set train/test data frames
            args['train_data'] = df.iloc[train_index, :]
            args['test_data'] = df.iloc[test_index, :]

            # Run benchmark
            recommendations, metrics = _bench(**args)

            # Append
            recommendations_list.append(recommendations)
            metrics_list.append(metrics)

            i += 1

        return recommendations_list, metrics_list


def _bench(recommenders: Dict[str, BanditRecommender],
           metrics: List[Union[BinaryRecoMetrics, RankingRecoMetrics]],
           train_data: Union[str, pd.DataFrame],
           test_data: Union[str, pd.DataFrame],
           user_features: Union[str, pd.DataFrame] = None,
           user_features_list: Union[str, List[str]] = None,
           user_features_dtypes: Union[str, Dict] = None,
           item_features: Union[str, pd.DataFrame] = None,
           item_list: Union[str, List[str]] = None,
           item_eligibility: Union[str, pd.DataFrame] = None,
           warm_start: bool = False,
           warm_start_distance: float = None,
           user_id_col: str = Constants.user_id,
           item_id_col: str = Constants.item_id,
           response_col: str = Constants.response,
           batch_size: int = 100000,
           verbose: bool = False) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Dict[str, float]]]:

    # Import data
    train_data_df, item_list, user_features_df, \
        item_to_features, _ = load_data(data=train_data,
                                        user_features=user_features,
                                        user_features_list=user_features_list,
                                        user_features_dtypes=user_features_dtypes,
                                        item_features=item_features,
                                        item_list=item_list,
                                        item_eligibility=item_eligibility,
                                        user_id_col=user_id_col,
                                        item_id_col=item_id_col,
                                        response_col=response_col)
    test_data_df = load_response_data(test_data)

    recommendations = dict()
    rec_metrics = dict()
    for name, recommender in recommenders.items():
        if verbose:
            print(f'>>> {name}')
            print('Running...')

        # Copy recommender
        rec = deepcopy(recommender)

        # Train
        t0 = time()
        train(
            recommender=rec,
            data=train_data_df,
            user_features=user_features_df,
            item_features=item_features,
            item_list=item_list,
            warm_start=warm_start,
            warm_start_distance=warm_start_distance,
            user_id_col=user_id_col,
            item_id_col=item_id_col,
            response_col=response_col,
            batch_size=batch_size,
            save_file=None)

        # Score
        recommendations[name] = score(
            recommender=rec,
            data=test_data_df,
            user_features=user_features_df,
            item_features=item_features,
            item_list=item_list,
            item_eligibility=item_eligibility,
            warm_start=warm_start,
            warm_start_distance=warm_start_distance,
            user_id_col=user_id_col,
            item_id_col=item_id_col,
            response_col=response_col,
            batch_size=batch_size,
            save_file=None)

        # Evaluate
        cm = CombinedMetrics(*metrics)
        rec_metrics[name] = cm.get_score(test_data_df.rename(columns={response_col: Constants.score}),
                                         recommendations[name])
        if verbose:
            print(f"Done: {(time() - t0) / 60:.2f} minutes \n")

    return recommendations, rec_metrics


def _validate_common_args(data, user_features, user_features_list, user_features_dtypes, item_features,
                          item_list, item_eligibility, warm_start, warm_start_distance, user_id_col, item_id_col,
                          response_col, batch_size):
    # Train/test data
    check_true(data is not None, ValueError("Data input cannot be none."))
    check_true(isinstance(data, (str, pd.DataFrame)),
               TypeError("Data should be string of filepath or data frame."))

    # User features
    if user_features is not None:
        check_true(isinstance(user_features, (str, pd.DataFrame)),
                   TypeError("User features should be string of filepath or data frame."))
    if user_features_list is not None:
        check_true(user_features is not None, ValueError("User features should be given if user list is specified."))
        check_true(isinstance(user_features_list, (str, list)),
                   TypeError("User features list should be string of filepath or list."))
    if user_features_dtypes is not None:
        check_true(user_features is not None, ValueError("User features should be given if user dtypes is specified."))
        check_true(isinstance(user_features_dtypes, (str, dict)),
                   TypeError("User features dtypes should be string of filepath or dictionary."))

    # Item features
    if item_features is not None:
        check_true(isinstance(item_features, (str, pd.DataFrame)),
                   TypeError("Item features should be string of filepath or data frame."))
    if item_list is not None:
        check_true(isinstance(item_list, (str, list)),
                   TypeError("Item list should be string of filepath or list."))
    if item_eligibility is not None:
        check_true(isinstance(item_eligibility, (str, pd.DataFrame)),
                   TypeError("Item eligibility should be string of filepath or data frame."))

    # Warm start
    check_true(isinstance(warm_start, bool), TypeError("Warm start flag should be boolean."))
    if warm_start:
        check_true(warm_start_distance is not None, ValueError("Warm start distance cannot be none."))
        check_true(isinstance(warm_start_distance, float), TypeError("Warm start distance should be a float."))
        check_true(isinstance(warm_start_distance, float), TypeError("Warm start distance should be a float."))
        check_true(0 <= warm_start_distance <= 1, ValueError("Warm start distance should be between 0 and 1."))
        check_true(item_features is not None, ValueError("Item features are required to warm start arms."))
    else:
        check_true(warm_start_distance is None, ValueError("Warm start distance should be none if warm start false."))

    # IDs
    check_true(isinstance(user_id_col, str), TypeError("User id should be a string."))
    check_true(isinstance(item_id_col, str), TypeError("Item id should be a string."))
    check_true(isinstance(response_col, str), TypeError("Response column should be a string."))

    # Batch size
    check_true(isinstance(batch_size, int), TypeError("Batch size should be an integer."))
    check_true(batch_size > 0, ValueError("Batch size should be positive."))


def _validate_recommender(recommender, is_fit=False):
    if not isinstance(recommender, dict):
        recommender_dict = {"": recommender}
    else:
        recommender_dict = recommender
    for rec in recommender_dict.values():
        check_true(isinstance(rec, BanditRecommender), TypeError("Recommender should be a BanditRecommender instance."))
        if is_fit:
            check_true(rec.mab is not None, ValueError("Recommender has not been initialized."))
            check_true(rec.mab._is_initial_fit, ValueError("Recommender has not been fit."))


def _validate_save(save_file):
    if save_file is not None:
        check_true(isinstance(save_file, (bool, str)), TypeError("Save file should be boolean or a string filepath."))


def _validate_bench(recommenders, metrics, train_data, test_data, cv):

    # Recommenders
    check_true(recommenders is not None, ValueError("Recommenders cannot be none."))
    check_true(isinstance(recommenders, dict), TypeError("Recommenders should be given as a dictionary."))

    # Metrics
    check_true(isinstance(metrics, list), TypeError("Metrics should be given as a list."))
    for v in metrics:
        check_true(isinstance(v, (BinaryRecoMetrics.AUC,
                                  BinaryRecoMetrics.CTR,
                                  RankingRecoMetrics.Precision,
                                  RankingRecoMetrics.Recall,
                                  RankingRecoMetrics.NDCG,
                                  RankingRecoMetrics.MAP)),
                   TypeError("Evaluation metric values must be BinaryRecoMetrics or RankingRecoMetrics instances."))

    # Train/test data
    check_true(train_data is not None, ValueError("Train data cannot be none."))
    check_true(isinstance(train_data, (str, pd.DataFrame)),
               TypeError("Train data should be string of filepath or data frame."))
    if test_data is not None:
        check_true(isinstance(test_data, (str, pd.DataFrame)),
                   TypeError("Test data should be string of filepath or data frame."))

    # CV
    if cv is not None:
        check_true(isinstance(cv, int), TypeError("Cross-validation (cv) must be an integer."))
        check_true(test_data is None, ValueError("Test data must be None when using Cross-validation (cv)."))



# -*- coding: utf-8 -*-
# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

import json
import pickle
from typing import Dict, List, NamedTuple, NoReturn, Union

import numpy as np
import pandas as pd
from jurity.recommenders import BinaryRecoMetrics, RankingRecoMetrics
from mabwiser.utils import Arm, Num
from mabwiser.utils import check_true


class Constants(NamedTuple):
    """
    Constant values used by the modules.
    """
    default_seed = 12345
    user_id = 'user_id'
    item_id = 'item_id'
    response = 'response'
    score = 'score'


def explode_recommendations(df: pd.DataFrame, unique_col: str, explode_cols: List[str]):
    """Replicates the explode functionality in pandas 0.25 and later.

    Assumes that there are two levels in the dataframe. The unique column is the first level, and it contains
    de-duplicated values. The columns in explode_cols is the second level, where each of these columns contain a list
    of values. Each column in explode_cols is assumed to contain a list of same length.

    The output is the normalized dataframe where the lists are split into individual rows.
    """
    # First, remove anything with 0 length
    lens = df[explode_cols[0]].str.len()
    df = df[lens != 0].reset_index(drop=True)
    # Calculate lengths of the second level list of values
    lens = df[explode_cols[0]].str.len()
    # Repeat the unique column to get the same number of values as the second level
    unique_vals = np.repeat(df[unique_col], lens.values)
    cols = {unique_col: unique_vals.values}
    # Concatenate all second level values to get a flattened list
    cols.update({key: np.concatenate(df[key]) for key in explode_cols})
    return pd.DataFrame(cols)


def concat_recommendations_list(recommendation_results_list: List[Dict[str, pd.DataFrame]]) -> Dict[str, pd.DataFrame]:
    """
    Concatenates recommendation results split across multiple data frames into a single dataframe.

    Parameters
    ----------
    recommendation_results_list: List[Dict[str, pd.DataFrame]]
        List of dictionaries returned by benchmark function.

    Returns
    -------
    Dictionary with recommendations by algorithm.
    """
    recommendation_results_concat = dict()
    for recommendation_results in recommendation_results_list:
        for name, df in recommendation_results.items():
            if name not in recommendation_results_concat:
                recommendation_results_concat[name] = df
            else:
                recommendation_results_concat[name] = pd.concat((recommendation_results_concat[name], df))
    return recommendation_results_concat


def default_metrics(top_k_values=None, user_id_col: str = Constants.user_id, item_id_col: str = Constants.item_id):
    metric_params = {'click_column': Constants.score, 'user_id_column': user_id_col, 'item_id_column': item_id_col}
    metrics = []
    for k in top_k_values:
        metrics.append(BinaryRecoMetrics.AUC(**metric_params, k=k))
        metrics.append(BinaryRecoMetrics.CTR(**metric_params, k=k))
        metrics.append(RankingRecoMetrics.Precision(**metric_params, k=k))
        metrics.append(RankingRecoMetrics.Recall(**metric_params, k=k))
        metrics.append(RankingRecoMetrics.NDCG(**metric_params, k=k))
        metrics.append(RankingRecoMetrics.MAP(**metric_params, k=k))
    return metrics


def load_data(data: Union[str, pd.DataFrame], user_features: Union[str, pd.DataFrame] = None,
              user_features_list: Union[str, List[str]] = None, user_features_dtypes: Union[str, Dict] = None,
              item_features: Union[str, pd.DataFrame] = None, item_list: Union[str, List[str]] = None,
              item_eligibility: Union[str, pd.DataFrame] = None, user_id_col: str = Constants.user_id,
              item_id_col: str = Constants.item_id, response_col: str = Constants.response):
    """
    Import data.

    Parameters
    ----------
    data: Union[str, pd.DataFrame]
        Data should have a row for each sample (user_id, item_id, response).
        Column names should be consistent with user_id_col, item_id_col and response_col arguments.
        CSV format with file header or Data Frame.
    user_features: Union[str, pd.DataFrame]
        User features containing features for each user_id.
        Each row should include user_id and list of features (user_id, u_1, u_2, ..., u_p).
        CSV format with file header or Data Frame.
    user_features_list: Union[str, List[str]]
        List of user features to use.
        Must be a subset of features in (u_1, u_2, ... u_p).
        If None, all the features in user_features are used.
        CSV format with file header or List.
    user_features_dtypes: Union[str, Dict]
        User features data types file with mappings of features to their dtypes upon loading.
        Data should have a key, value pair for user feature, e.g., {"feature_1": "float32"}
        The keys should be consistent with `user_features` file.
    user_id_col : str, default=Constants.user_id
        User id column name.
    item_features: Union[str, pd.DataFrame]
        Item features file containing features for each item_id.
        Each row should include item_id and list of features (item_id, i_1, i_2, .... i_q).
        CSV format with file header or Data Frame.
    item_eligibility: Union[str, pd.DataFrame], default=None
        Items each user is eligible for.
        Used to generate excluded_arms lists.
        If None, all the items can be evaluated for recommendation for each user.
        CSV format with file header or Data Frame.
    item_list: List[Arm]
        List of items.
    item_id_col : str, default=Constants.item_id
        Item id column name.
    response_col : str, default=Constants.response
        Response column name.

    Returns
    -------
    Data frame with response data.
    """

    # Response data
    data_df = load_response_data(data, user_id_col, item_id_col, response_col)

    # Item list
    item_list = load_items(data_df, item_list, item_id_col)

    # User features
    if user_features is not None:
        user_features_df = load_user_features(user_features, user_features_list, user_features_dtypes, user_id_col)
    else:
        user_features_df = None

    # Item features
    if item_features is not None:
        item_to_features = load_item_features(item_features, item_list, item_id_col)
    else:
        item_to_features = None

    # Item eligibility
    if item_eligibility is not None:
        excluded_df = load_excluded_items(item_eligibility, item_list, user_id_col, item_id_col)
    else:
        excluded_df = None

    return data_df, item_list, user_features_df, item_to_features, excluded_df


def load_response_data(data: Union[str, pd.DataFrame], user_id_col: str = Constants.user_id,
                       item_id_col: str = Constants.item_id, response_col: str = Constants.response) -> pd.DataFrame:
    """
    Import response data.

    Parameters
    ----------
    data: Union[str, pd.DataFrame]
        Data should have a row for each sample (user_id, item_id, response).
        Column names should be consistent with user_id_col, item_id_col and response_col arguments.
        CSV format with file header or Data Frame.
    user_id_col : str, default=Constants.user_id
        User id column name.
    item_id_col : str, default=Constants.item_id
        Item id column name.
    response_col : str, default=Constants.response
        Response column name.

    Returns
    -------
    Data frame with response data.
    """
    df = load_data_frame(data)
    check_true(user_id_col in df.columns, ValueError(f"{user_id_col} not in data file."))
    check_true(item_id_col in df.columns, ValueError(f"{item_id_col} not in data file."))
    check_true(response_col in df.columns, ValueError(f"{response_col} not in data file."))
    return df[[user_id_col, item_id_col, response_col]].astype({response_col: int})


def load_items(data_df: pd.DataFrame, item_list: Union[str, List[str]] = None, item_id_col: str = Constants.item_id):
    """
    Import item list.

    Parameters
    ----------
    data_df: pd.DataFrame
        Data frame with response data.
        Data should have a row for each sample (user_id, item_id, response).
    item_list: Union[str, List[str]], default=None
        List of items.
        If None, the list of items in data_df are returned.
        CSV format with file header or List.
    item_id_col : str, default=Constants.item_id
        Item id column name.

    Returns
    -------
    List of items.
    """
    if item_list is None:
        check_true(item_id_col in data_df.columns, ValueError(f"{item_id_col} not in data file."))
        return data_df[item_id_col].unique().tolist()
    else:
        return load_list(item_list)


def load_item_features(item_features: Union[str, pd.DataFrame], item_list: List[Arm],
                       item_id_col: str = Constants.item_id) -> Dict[Arm, List[Num]]:
    """
    Import item features.

    Parameters
    ----------
    item_features: Union[str, pd.DataFrame]
        Item features file containing features for each item_id.
        Each row should include item_id and list of features (item_id, i_1, i_2, .... i_q).
        CSV format with file header or Data Frame.
    item_list: List[Arm]
        List of items.
    item_id_col: str
        Item id column name.

    Returns
    -------
    Dictionary mapping item features to each item.
    """
    df = load_data_frame(item_features)
    check_true(item_id_col in df.columns, ValueError(f"{item_id_col} not in item features."))
    check_true(len(df) == df[item_id_col].nunique(), ValueError(f"Duplicate item ids in item features."))

    # Convert from data frame to dictionary
    item_to_features = df.set_index(item_id_col).T.to_dict("list")

    # Drop features for items not in item list
    item_to_features = {item: features for item, features in item_to_features.items() if item in item_list}

    # Raise error if features are missing for item in item list
    for item in item_list:
        if item not in item_to_features:
            raise ValueError(f"{item} not found in item features.")

    return item_to_features


def load_user_features(user_features: Union[str, pd.DataFrame], user_features_list: Union[str, List[str]] = None,
                       user_features_dtypes: Union[str, Dict] = None,
                       user_id_col: str = Constants.user_id) -> pd.DataFrame:
    """
    Import user features.

    Parameters
    ----------
    user_features: Union[str, pd.DataFrame]
        User features containing features for each user_id.
        Each row should include user_id and list of features (user_id, u_1, u_2, ..., u_p).
        CSV format with file header or Data Frame.
    user_features_list: Union[str, List[str]]
        List of user features to use.
        Must be a subset of features in (u_1, u_2, ... u_p).
        If None, all the features in user_features are used.
        CSV format with file header or List.
    user_features_dtypes: Union[str, Dict]
        User features data types file with mappings of features to their dtypes upon loading.
        Data should have a key, value pair for user feature, e.g., {"feature_1": "float32"}
        The keys should be consistent with `user_features` file.
    user_id_col: str
        User id column name.

    Returns
    -------
    Data frame with user features.
    """
    # User features data types
    data_types = None
    if user_features_dtypes is not None:
        if isinstance(user_features_dtypes, str):
            with open(user_features_dtypes, 'r') as f:
                data_types = json.load(f)
        else:
            data_types = user_features_dtypes

    # Load data
    if isinstance(user_features, str):
        df = pd.read_csv(user_features, dtype=data_types)
    else:
        df = pd.DataFrame(user_features)
    check_true(user_id_col in df.columns, ValueError(f"{user_id_col} not in user features."))
    check_true(len(df) == df[user_id_col].nunique(), ValueError(f"Duplicate user ids in user features."))

    # Subset
    if user_features_list is not None:
        keep = [user_id_col] + load_list(user_features_list)
        return df[keep]
    else:
        return df


def load_data_frame(data: Union[str, pd.DataFrame]) -> pd.DataFrame:
    """
    Load file as data frame.

    Parameters
    ----------
    data: Union[str, pd.DataFrame]
        CSV file with header if string input, otherwise Data Frame.

    Returns
    -------
    Data frame with user features.
    """
    if isinstance(data, str):
        return pd.read_csv(data)
    elif isinstance(data, pd.DataFrame):
        return data
    else:
        raise TypeError("Data must be string of filepath or data frame.")


def load_excluded_items(item_eligibility: Union[str, pd.DataFrame], item_list: List[Arm],
                        user_id_col: str = Constants.user_id, item_id_col: str = Constants.item_id) -> pd.DataFrame:
    """Convert eligibility data to excluded_arms list for each user_id.

    Parameters
    ----------
    item_eligibility: Union[str, pd.DataFrame], default=None
        Items each user is eligible for.
        Used to generate excluded_arms lists.
        If None, all the items can be evaluated for recommendation for each user.
        CSV format with file header or Data Frame.
    item_list: List
        The list of all arms.
    user_id_col: str
        User id column name.
    item_id_col: str
        Item id column name.

    Returns
    -------
    DataFrame with user id and list of list of arms to exclude from recommended arms.
    """

    # Load data
    df = load_data_frame(item_eligibility).copy()
    check_true(user_id_col in df.columns, ValueError(user_id_col + ' missing from eligibility_data.'))
    check_true(item_id_col in df.columns, ValueError(item_id_col + ' missing from eligibility_data.'))

    # Create list of excluded items for each user.
    df['excluded_arms'] = df.apply(lambda x: get_exclusion_list(item_list, x[item_id_col]), axis=1)
    df.drop(item_id_col, axis=1, inplace=True)
    df.columns = [user_id_col, item_id_col]
    return df


def load_pickle(pickle_file: str):
    """
    Returns the loaded pickle object.
    """
    with open(pickle_file, 'rb') as infile:
        return pickle.load(infile)


def load_list(data: Union[str, List]) -> List:
    """
    Load file as list.

    Parameters
    ----------
    data: Union[str, pd.DataFrame]
        CSV file with header if string input, otherwise List.

    Returns
    -------
    Data frame with user features.
    """
    if isinstance(data, str):
        return pd.read_csv(data).iloc[:, 0].tolist()
    elif isinstance(data, list):
        return data
    else:
        raise TypeError("Data must be string of filepath or list.")


def get_exclusion_list(arms, eligible_list):
    return list(set(arms).difference(set(eval(eligible_list))))


def print_interaction_stats(df: pd.DataFrame, user_id_col: str = Constants.user_id,
                            item_id_col: str = Constants.item_id, response_col: str = Constants.response) -> NoReturn:
    """
    Print number of rows, number of users, number of items in interaction data.

    Parameters
    ----------
    df: pd.DataFrame
        Interaction data frame with (user_id, item_id, response) in each row.
    user_id_col: str
        User id column name.
    item_id_col: str
        Item id column name.
    response_col: str
        Response column name.

    Returns
    -------
    No return.
    """

    print(f"Number of rows: {len(df):,}")
    print(f"Number of users: {df[user_id_col].nunique():,}")
    print(f"Number of items: {df[item_id_col].nunique():,}")
    print(f"Mean response rate: {df[response_col].mean():.4f}\n")


def merge_user_features(responses_df: pd.DataFrame, user_features_df: pd.DataFrame,
                        user_id_col: str = Constants.user_id) -> pd.DataFrame:
    """
    Merge responses and user features.

    Parameters
    ----------
    responses_df : pd.DataFrame
        Responses.
    user_features_df : pd.DataFrame
        User features.
    user_id_col: str
        User id column name.

    Returns
    -------
    Data frame with merged responses and user features.
    """
    # Subset features to only include users in response data and then merge
    df = user_features_df[user_features_df[user_id_col].isin(responses_df[user_id_col].unique())]
    return responses_df.merge(df, on=user_id_col, how="left")


def save_json(obj, json_file) -> NoReturn:
    """
    Save obj as json file.
    """
    with open(json_file, 'w') as f:
        json.dump(obj, f)


def save_pickle(obj, pickle_file) -> NoReturn:
    """
    Save serializable object as pickle file.
    """
    with open(pickle_file, 'wb') as fp:
        pickle.dump(obj, fp)



# -*- coding: utf-8 -*-
# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

import os

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    required = fh.read().splitlines()

with open(os.path.join('mab2rec', '_version.py')) as fp:
    exec(fp.read())

setuptools.setup(
    name="mab2rec",
    description="Mab2Rec: Multi-Armed Bandits Recommender",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=__version__,
    author=__author__,
    url="",
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=required,
    python_requires=">=3.7",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Source": "https://github.com/fidelity/mab2rec"
    }
)



# -*- coding: utf-8 -*-
# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

from enum import Enum
from typing import Dict, Optional

from spock import SpockBuilder
from spock.config import spock

from mab2rec import BanditRecommender, LearningPolicy, NeighborhoodPolicy
from mab2rec.utils import Constants


@spock
class EpsilonGreedy:
    """Epsilon Greedy Learning Policy.

    Attributes
    ----------
    epsilon: Num
        The probability of selecting a random arm for exploration.
        Integer or float. Must be between 0 and 1.
        Default value is 0.1.
    """
    epsilon: float = 0.1


@spock
class LinGreedy:
    """LinGreedy Learning Policy

    Attributes
    ----------
    epsilon: Num
        The probability of selecting a random arm for exploration.
        Integer or float. Must be between 0 and 1.
        Default value is 0.1.
    l2_lambda: Num
        The regularization strength.
        Integer or float. Must be greater than zero.
        Default value is 1.0.
    scale: bool
        Whether to scale features to have zero mean and unit variance.
        Uses StandardScaler in sklearn.preprocessing.
        Default value is False.
    """
    epsilon: float = 0.1
    l2_lambda: float = 1.0
    scale: bool = False


@spock
class LinTS:
    """LinTS Learning Policy

    Attributes
    ----------
    alpha: Num
        The multiplier to determine the degree of exploration.
        Integer or float. Must be greater than zero.
        Default value is 1.0.
    l2_lambda: Num
        The regularization strength.
        Integer or float. Must be greater than zero.
        Default value is 1.0.
    scale: bool
        Whether to scale features to have zero mean and unit variance.
        Uses StandardScaler in sklearn.preprocessing.
        Default value is False.
    """
    alpha: float = 1.0
    l2_lambda: float = 1.0
    scale: bool = False


@spock
class LinUCB:
    """LinUCB Learning Policy.

    Attributes
    ----------
    alpha: Num
        The parameter to control the exploration.
        Integer or float. Cannot be negative.
        Default value is 1.0.
    l2_lambda: Num
        The regularization strength.
        Integer or float. Cannot be negative.
        Default value is 1.0.
    scale: bool
        Whether to scale features to have zero mean and unit variance.
        Uses StandardScaler in sklearn.preprocessing.
        Default value is False.
    """
    alpha: float = 1.0
    l2_lambda: float = 1.0
    scale: bool = False


@spock
class Popularity:
    """Randomized Popularity Learning Policy."""
    none: Optional[str] = None


@spock
class Random:
    """Random Learning Policy."""
    none: Optional[str] = None


@spock
class Softmax:
    """Softmax Learning Policy.

      Attributes
      ----------
      tau: Num
           The temperature to control the exploration.
           Integer or float. Must be greater than zero.
           Default value is 1.
    """
    tau: float = 1.0


@spock
class ThompsonSampling:
    """Thompson Sampling Learning Policy."""
    none: Optional[str] = None


@spock
class UCB1:
    """Upper Confidence Bound1 Learning Policy.

      Attributes
      ----------
      alpha: Num
          The parameter to control the exploration.
          Integer of float. Cannot be negative.
          Default value is 1.
    """
    alpha: float = 1.0


@spock
class Clusters:
    """Clusters Neighborhood Policy.

    Attributes
    ----------
    n_clusters: Num
        The number of clusters. Integer. Must be at least 2. Default value is 2.
    is_minibatch: bool
        Boolean flag to use ``MiniBatchKMeans`` or not. Default value is False.
    """
    n_clusters: int = 2
    is_minibatch: bool = False


@spock
class KNearest:
    """KNearest Neighborhood Policy.

    Attributes
    ----------
    k: int
        The number of neighbors to select.
        Integer value. Must be greater than zero.
        Default value is 1.
    metric: str
        The metric used to calculate distance.
        Accepts any of the metrics supported by ``scipy.spatial.distance.cdist``.
        Default value is Euclidean distance.
    """
    k: int = 1
    metric: str = "euclidean"


@spock
class LSHNearest:
    """Locality-Sensitive Hashing Approximate Nearest Neighbors Policy.

    Attributes
    ----------
    n_dimensions: int
        The number of dimensions to use for the hyperplane.
        Integer value. Must be greater than zero.
        Default value is 5.
    n_tables: int
        The number of hash tables.
        Integer value. Must be greater than zero.
        Default value is 3.
    """
    n_dimensions: int = 5
    n_tables: int = 3


@spock
class Radius:
    """Radius Neighborhood Policy.

    Radius is a nearest neighborhood approach that selects the observations
    within a given *radius* to be used with a learning policy.

    Attributes
    ----------
    radius: Num
        The maximum distance within which to select observations.
        Integer or Float. Must be greater than zero.
        Default value is 1.
    metric: str
        The metric used to calculate distance.
        Accepts any of the metrics supported by scipy.spatial.distance.cdist.
        Default value is Euclidean distance.
    """
    radius: float = 0.05
    metric: str = "euclidean"


@spock
class TreeBandit:
    """TreeBandit Neighborhood Policy.

    Attributes
    ----------
    tree_parameters: Dict, **kwarg
        Parameters of the decision tree.
        The keys must match the parameters of sklearn.tree.DecisionTreeClassifier.
        When a parameter is not given, the default parameters from
        sklearn.tree.DecisionTreeClassifier will be chosen.
        Default value is an empty dictionary.
    """
    tree_parameters: Dict = {}


class LearningPolicyOptions(Enum):
    """Learning Policies
    """
    epsilon_greedy = EpsilonGreedy
    lin_greedy = LinGreedy
    lin_ts = LinTS
    lin_ucb = LinUCB
    popularity = Popularity
    random = Random
    softmax = Softmax
    ts = ThompsonSampling
    ucb1 = UCB1


class NeighborhoodPolicyOptions(Enum):
    """Neighborhood Policies"""
    clusters = Clusters
    knearest = KNearest
    lshnearest = LSHNearest
    radius = Radius
    treebandit = TreeBandit


@spock
class Recommender:
    learning_policy: LearningPolicyOptions
    neighborhood_policy: Optional[NeighborhoodPolicyOptions] = None
    top_k: Optional[int] = 10
    seed: Optional[int] = 123456
    n_jobs: Optional[int] = 1
    backend: Optional[str] = None


@spock
class RecommenderScore:
    recommender_pickle: str


@spock
class DataConfig:
    data: str
    user_features: Optional[str]
    user_features_list: Optional[str]
    user_features_dtypes: Optional[str]
    item_features: Optional[str]
    item_list: Optional[str]
    item_eligibility: Optional[str]
    warm_start: Optional[bool] = False
    warm_start_distance: Optional[float] = None
    user_id_col: Optional[str] = Constants.user_id
    item_id_col: Optional[str] = Constants.item_id
    response_col: Optional[str] = Constants.response
    batch_size: Optional[int] = 100000
    save_file: Optional[str] = None


def init_recommender(config):

    # Learning policy
    lp_params = config.learning_policy
    if isinstance(lp_params, EpsilonGreedy):
        lp = LearningPolicy.EpsilonGreedy(epsilon=lp_params.epsilon)
    elif isinstance(lp_params, LinGreedy):
        lp = LearningPolicy.LinGreedy(epsilon=lp_params.epsilon, l2_lambda=lp_params.l2_lambda, scale=lp_params.scale)
    elif isinstance(lp_params, LinTS):
        lp = LearningPolicy.LinTS(alpha=lp_params.alpha, l2_lambda=lp_params.l2_lambda, scale=lp_params.scale)
    elif isinstance(lp_params, LinUCB):
        lp = LearningPolicy.LinUCB(alpha=lp_params.alpha, l2_lambda=lp_params.l2_lambda, scale=lp_params.scale)
    elif isinstance(lp_params, Popularity):
        lp = LearningPolicy.Popularity()
    elif isinstance(lp_params, Random):
        lp = LearningPolicy.Random()
    elif isinstance(lp_params, Softmax):
        lp = LearningPolicy.Softmax(tau=lp_params.tau)
    elif isinstance(lp_params, ThompsonSampling):
        lp = LearningPolicy.ThompsonSampling()
    elif isinstance(lp_params, UCB1):
        lp = LearningPolicy.UCB1(alpha=lp_params.alpha)
    else:
        raise NotImplemented(f'{lp_params} not implemented')

    # Neighborhood policy
    cp_params = config.neighborhood_policy
    if isinstance(lp_params, Clusters):
        cp = NeighborhoodPolicy.Clusters(n_clusters=cp_params.n_clusters, is_minibatch=cp_params.is_minibatch)
    elif isinstance(lp_params, KNearest):
        cp = NeighborhoodPolicy.KNearest(k=cp_params.k, metric=cp_params.metric)
    elif isinstance(lp_params, LSHNearest):
        cp = NeighborhoodPolicy.LSHNearest(n_dimensions=cp_params.n_dimensions, n_tables=cp_params.n_tables)
    elif isinstance(lp_params, Radius):
        cp = NeighborhoodPolicy.Radius(radius=cp_params.radius, metric=cp_params.metric)
    elif isinstance(lp_params, TreeBandit):
        cp = NeighborhoodPolicy.TreeBandit(tree_parameters=cp_params.tree_parameters)
    elif cp_params is None:
        cp = None
    else:
        raise NotImplemented(f'{cp_params} not implemented')

    return BanditRecommender(lp, cp, config.top_k, config.seed, config.n_jobs, config.backend)


def train_config():

    config = SpockBuilder(DataConfig, Recommender,
                          EpsilonGreedy, LinGreedy, LinTS, LinUCB, Popularity, Random, Softmax, ThompsonSampling, UCB1,
                          Clusters, KNearest, LSHNearest, Radius, TreeBandit,
                          desc='Train Recommender').generate()

    data_config = config.DataConfig
    rec = init_recommender(config.Recommender)
    args = {
        'recommender': rec,
        'data': data_config.data,
        'user_features': data_config.user_features,
        'user_features_list': data_config.user_features_list,
        'user_features_dtypes': data_config.user_features_dtypes,
        'item_features': data_config.item_features,
        'item_list': data_config.item_list,
        'item_eligibility': data_config.item_eligibility,
        'warm_start': data_config.warm_start,
        'warm_start_distance': data_config.warm_start_distance,
        'user_id_col': data_config.user_id_col,
        'item_id_col': data_config.item_id_col,
        'response_col': data_config.response_col,
        'batch_size': data_config.batch_size,
        'save_file': data_config.save_file,
    }
    return args


def score_config():

    config = SpockBuilder(DataConfig, RecommenderScore, desc='Score Recommender').generate()

    data_config = config.DataConfig
    args = {
        'recommender': config.RecommenderScore.recommender_pickle,
        'data': data_config.data,
        'user_features': data_config.user_features,
        'user_features_list': data_config.user_features_list,
        'user_features_dtypes': data_config.user_features_dtypes,
        'item_features': data_config.item_features,
        'item_list': data_config.item_list,
        'item_eligibility': data_config.item_eligibility,
        'warm_start': data_config.warm_start,
        'warm_start_distance': data_config.warm_start_distance,
        'user_id_col': data_config.user_id_col,
        'item_id_col': data_config.item_id_col,
        'response_col': data_config.response_col,
        'batch_size': data_config.batch_size,
        'save_file': data_config.save_file,
    }
    return args
